from shutil import copy2, copytree
from os import listdir, makedirs
from os.path import join, isdir, isfile, dirname
from yaml import load as yaml_load, SafeLoader
from tiramisu.error import PropertiesOptionError
#
from .utils import RISOTTO_CONFIG


class ModuleCfg():
    def __init__(self, module_name):
        self.module_name = module_name
        self.dictionaries_dir = []
        self.functions_file = []
        self.templates_dir = []
        self.patches_dir = []
        self.extra_dictionaries = {}
        self.servers = []
        self.depends = []
        self.manuals = []
        self.tests = []
        self.providers = []
        self.suppliers = []

    def __repr__(self):
        return str(vars(self))

def load_application_service(as_dir: str) -> str:
    with open(join(as_dir, 'applicationservice.yml')) as yaml:
        return yaml_load(yaml, Loader=SafeLoader)

class Applications:
    def __init__(self) -> None:
        self.datasets = RISOTTO_CONFIG.get('directories', {}).get('datasets', ['dataset'])
        self.application_directories = self._load_application_directories()

    def _load_application_directories(self) -> dict:
        """List all service applications in datasets
        Returns something link:
        {<applicationservice>: seed/<applicationservice>}
        """
        applications = {'host': None}
        for dataset_directory in self.datasets:
            for applicationservice in listdir(dataset_directory):
                applicationservice_dir = join(dataset_directory, applicationservice)
                if not isdir(applicationservice_dir) or \
                        not isfile(join(applicationservice_dir, 'applicationservice.yml')):
                    continue
                if applicationservice in applications:
                    raise Exception(f'multi applicationservice: {applicationservice} ({applicationservice_dir} <=> {applications[applicationservice]})')
                applications[applicationservice] = applicationservice_dir
        return applications


class Modules:
    """Modules are defined by the end user
    A module is the a list of service applications
    The class collects all the useful information for the module
    """

    def __init__(self,
                 applicationservices: Applications,
                 applicationservice_provider: str,
                 modules_name: list,
                 host_applicationsservice: str,
                 ) -> None:
        self.application_directories = applicationservices.application_directories
        self.module_infos = {}
        self.module_infos['host'] = self._load_module_informations('host',
                                                                   ['host', host_applicationsservice],
                                                                   is_host=True,
                                                                   )
        for module_name in modules_name:
            if module_name == 'host':
                raise Exception('forbidden module name: "host"')
            self.module_infos[module_name] = self._load_module_informations(module_name,
                                                                            [applicationservice_provider, module_name],
                                                                            is_host=False,
                                                                            )

    def get(self,
            module_name: str,
            ) -> ModuleCfg:
        return self.module_infos[module_name]

    def _load_module_informations(self,
                                  module_name: str,
                                  applicationservices: list,
                                  is_host: bool,
                                  ) -> ModuleCfg:
        """Create a ModuleCfg object and collect informations
        A module must depend to an unique distribution
        """
        cfg = ModuleCfg(module_name)
        distribution = None

        for applicationservice in applicationservices:
            ret = self._load_applicationservice(applicationservice,
                                                cfg,
                                                )
            if ret:
                if distribution:
                    raise Exception(f'duplicate distribution for {cfg.module_name}: {distribution} and {ret} (dependencies: {cfg.depends}) ')
                distribution = ret
        if not is_host and not distribution:
            raise Exception(f'cannot found any linux distribution for {module_name}')
        return cfg

    def _load_applicationservice(self,
                                 appname: str,
                                 cfg: ModuleCfg,
                                 ) -> str:
        """extract informations from an application service and load it's dependency
        informations collected is store to the module

        returns the name of current distribution, if found
        """
        if appname not in self.application_directories:
            raise Exception(f'cannot find application dependency "{appname}"')
        cfg.depends.append(appname)
        as_dir = self.application_directories[appname]
        if not as_dir:
            return
        self._load_applicationservice_directories(as_dir,
                                                  cfg,
                                                  )
        app = load_application_service(as_dir)
        provider = app.get('provider')
        if provider:
            cfg.providers.setdefault(provider, [])
            if appname not in cfg.providers[provider]:
                cfg.providers[provider].append(appname)
        supplier = app.get('supplier')
        if supplier:
            self.suppliers.setdefault(supplier, [])
            if appname not in self.suppliers[supplier]:
                self.suppliers[supplier].append(appname)
        if 'distribution' in app and app['distribution']:
            distribution = appname
        else:
            distribution = None
        for depend in app.get('depends', []):
            if depend in cfg.depends:
                #this dependancy is already loaded for this module
                continue
            ret = self._load_applicationservice(depend,
                                                cfg,
                                                )
            if ret:
                if distribution:
                    raise Exception(f'duplicate distribution for {cfg.module_name}: {distribution} and {ret} (dependencies: {cfg.depends}) ')
                distribution = ret
        return distribution

    def _load_applicationservice_directories(self,
                                             as_dir: str,
                                             cfg: ModuleCfg,
                                             ) -> None:
        # dictionaries
        dictionaries_dir = join(as_dir, 'dictionaries')
        if isdir(dictionaries_dir):
            cfg.dictionaries_dir.append(dictionaries_dir)
        # funcs
        funcs_dir = join(as_dir, 'funcs')
        if isdir(funcs_dir):
            for f in listdir(funcs_dir):
                if f.startswith('__'):
                    continue
                cfg.functions_file.append(join(funcs_dir, f))
        # templates
        templates_dir = join(as_dir, 'templates')
        if isdir(templates_dir):
            cfg.templates_dir.append(templates_dir)
        # patches
        patches_dir = join(as_dir, 'patches')
        if isdir(patches_dir):
            cfg.patches_dir.append(patches_dir)
        # extras
        extras_dir = join(as_dir, 'extras')
        if isdir(extras_dir):
            for extra in listdir(extras_dir):
                extra_dir = join(extras_dir, extra)
                if isdir(extra_dir):
                    cfg.extra_dictionaries.setdefault(extra, []).append(extra_dir)
        # manual
        manual_dir = join(as_dir, 'manual', 'image')
        if isdir(manual_dir):
            cfg.manuals.append(manual_dir)
        # tests
        tests_dir = join(as_dir, 'tests')
        if isdir(tests_dir):
            cfg.tests.append(tests_dir)


def applicationservice_copy(src_file: str,
                            dst_file: str,
                            ) -> None:
    if isdir(src_file):
        if not isdir(dst_file):
            makedirs(dst_file)
        for subfilename in listdir(src_file):
            #if not copy_if_not_exists or not isfile(dst_file):
            src = join(src_file, subfilename)
            dst = join(dst_file, subfilename)
            if isfile(src):
                copy2(src, dst)
            else:
                copytree(src, dst)
    else:
        dst = dirname(dst_file)
        if not isdir(dst):
            makedirs(dst)
        if isfile(src_file):
            copy2(src_file, dst_file)
        else:
            copytree(src_file, dst_file)


async def valid_mandatories(config):
    mandatories = await config.value.mandatory()
    await config.property.pop('mandatory')
    hidden = {}
    variables = {}
    title = None
    if mandatories:
        server_name = None
        for mandatory in mandatories:
            path_server_name, path = mandatory.split('.', 1)
            var_server_name = await config.option(path_server_name).option.description()
            if server_name != var_server_name:
                server_name = var_server_name
                title = f'=== Missing variables for {server_name} ==='
            suboption = config.option(mandatory)
            text = await suboption.option.doc()
            msg = f' - {text} ({path})'
            supplier = await suboption.information.get('supplier', None)
            if supplier:
                msg += f' you could add a service that provides "{supplier}"'
            try:
                await config.option(mandatory).value.get()
                variables.setdefault(title, []).append(msg)
            except PropertiesOptionError as err:
                if 'hidden' not in err.proptype:
                    raise PropertiesOptionError(err)
                hidden.setdefault(title, []).append(msg)
        if not variables:
            variables = hidden
        return variables
